/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2005 Robert Ham <rah@bash.sh>
 *
 * GPL all the way, baby :)
 *
 */

#include <cmath>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_sf_hyperg.h>
#include <gsl/gsl_sf_result.h>
#include <gsl/gsl_sf_gamma.h>

#include "libtkcommon/debug.h"
#include "analyser.h"

using namespace TkResolver;
using namespace Analysis;

const Rational*
PositionFunction:: operator()(const Position &pos) const
{
  PositionFunction::const_iterator pos_iter( find(pos) );
  if (pos_iter == end())
    return 0;
  
  return &pos_iter->second;
}

Rational
PositionFunction:: calc_rss(const Probability &prob) const
{
  Rational sample_sum(0);
  for (PositionFunction::const_iterator pos_iter( begin() );
       pos_iter != end();
       ++pos_iter)
    sample_sum += pos_iter->second;
  
  Rational sample_mean( sample_sum / static_cast<Rational>(size()) );
  
  
  Rational difference;
  Rational square;
  Rational SS_E(0);
  for (PositionFunction::const_iterator pos_iter( begin() );
       pos_iter != end();
       ++pos_iter)
    {
      difference = pos_iter->second - prob;
      square = difference * difference;
      SS_E += square;
    }
  
  return SS_E;
}

Rational
PositionFunction:: calc_mean() const
{
  Rational sum(0);
  for (PositionFunction::const_iterator iter( begin() );
       iter != end();
       ++iter)
    sum += iter->second;
  
  return sum / static_cast<Rational>( size() );
}

Rational
PositionFunction:: calc_variance(const Rational &mean) const
{
  Rational sum(0);
  Rational difference;
  Rational square;
  
  for (PositionFunction::const_iterator iter( begin() );
       iter != end();
       ++iter)
    {
      difference = iter->second - mean;
      square = difference * difference;
      sum += square;
    }
  return sum / static_cast<Rational>( size() );
}

Real
PositionFunction:: calc_skewness(const Rational &mean, const Rational &variance) const
{
  Rational sum(0);
  Rational difference;
  Rational cube;
  
  for (PositionFunction::const_iterator iter( begin() );
       iter != end();
       ++iter)
    {
      difference = iter->second - mean;
      cube = difference * difference * difference;
      sum += cube;
    }
  return ( sqrt( static_cast<Real>(size()) ) * static_cast<Real>(sum) ) / pow(variance.get_d(), 2.0/3.0);
}

OffsetStatisticsSet:: OffsetStatisticsSet(const OffsetProbabilities &probs,
					  const PositionOffsetDistributions &offset_distributions,
					  const PositionCounts &position_counts)
  : _position_totals(position_counts)
{
  /* Offset list and theoretical probabilities */
  for (OffsetProbabilities::const_iterator prob_iter( probs.begin() );
       prob_iter != probs.end();
       ++prob_iter)
    {
      _offsets.push_back(prob_iter->first);
      _offset_stats[prob_iter->first]._probability = prob_iter->second;
    }
  
  /* Distributions and counts*/
  OffsetDistribution offset_position_counts;
  for (PositionCounts::const_iterator pos_iter( position_counts.begin() );
       pos_iter != position_counts.end();
       ++pos_iter)
    {
      const OffsetDistribution &offset_distribution( offset_distributions.find(pos_iter->first)->second );
      for (OffsetDistribution::const_iterator offset_iter( offset_distribution.begin() );
	   offset_iter != offset_distribution.end();
	   ++offset_iter)
	{
	  /* Distributions */
	  Rational div(static_cast<Rational>(offset_iter->second) / pos_iter->second);
	  PositionFunction::value_type pos_func_row(pos_iter->first, div);
	  _offset_stats[offset_iter->first]._position_function.insert(pos_func_row);
	  
	  /* Counts */
	  _offset_stats[offset_iter->first]._position_counts[pos_iter->first] = offset_iter->second;
	}
    }
  
  /* RSS, mean and variance */
  for (OffsetStatisticsMap::iterator offset_iter( _offset_stats.begin() );
       offset_iter != _offset_stats.end();
       ++offset_iter)
    {
      const PositionFunction &pos_func( offset_iter->second._position_function );
      OffsetStatistics &stats( offset_iter->second );
      stats._rss = pos_func.calc_rss(stats._probability);
      stats._mean = pos_func.calc_mean();
      stats._variance = pos_func.calc_variance(stats._mean);
      stats._skewness = pos_func.calc_skewness(stats._mean, stats._variance);
    }
}

Statistics:: Statistics(Db::Database &database,
			const OffsetProbabilities &offset_probs,
			const Db::IdList &experiments)
  : _n(0),
    _position_min(0),
    _position_max(0),
    _experiment_count(0)
{ 
  Offset results_row;
  Offset offset;
  int screen_pos;
  Real screen_width;
  Position pos;
  
  PositionCounts pos_counts;
  
  
  PositionOffsetDistributions pos_offsets;
  
  
  /*
   * Compile stats
   */
  for (Db::IdList::const_iterator ex_iter(experiments.begin());
       ex_iter != experiments.end();
       ++ex_iter)
    {
      ExperimentStatistics ex;
      Count ex_n(0);
      Position ex_pos_sum(0);
      
      ++_experiment_count;
      
      screen_pos = 0;
      pos = 0;
      screen_width = database.get_experiment_screen_width(*ex_iter);
     
      Db::DataSetPtr results( database.get_results(*ex_iter) );
      for((*results)->next();
	  (*results)->isAfterLast();
	  (*results)->next())
	{
	  offset = (*results)->getInt(1);
	  
	  screen_pos += offset;
	  pos = static_cast<Position>(screen_pos) / static_cast<Position>(screen_width);
	  
	  /* overall */
	  ++_n;
	  if (_position_max < pos)
	    _position_max = pos;
	  else if (_position_min > pos)
	    _position_min = pos;
	  
	  /* offset */
	  ++(pos_counts[pos]);
	  ++(pos_offsets[pos][offset]);
	  
	  /* experiment */
	  ++ex_n;
	  ex_pos_sum += pos;
	}
      
      /* complete experiment stats */
      ex._position_mean = ex_pos_sum / ex_n;
      _experiments.push_back(ex);
    }
  
  /* complete overall stats */
  _offsets = OffsetStatisticsSet(offset_probs,
				 pos_offsets,
				 pos_counts);
  
  /*
   * Calculate mean and variance of experiments' position mean
   */
  Position square;
  Position difference;
  
  /* grand mean */
  Position ex_pos_sum(0);
  for(ExperimentStatisticsSet::iterator iter(_experiments.begin());
      iter != _experiments.end();
      ++iter)
    ex_pos_sum += iter->_position_mean;
  
  _experiment_position_mean = ex_pos_sum / static_cast<Rational>(_experiment_count);
  
  /* variance */
  ex_pos_sum = 0;
  for(ExperimentStatisticsSet::iterator iter(_experiments.begin());
      iter != _experiments.end();
      ++iter)
    {
      difference = iter->_position_mean - _experiment_position_mean;
      square = difference * difference;
      ex_pos_sum += square;
    }
  
  _experiment_position_variance = ex_pos_sum / static_cast<Rational>(_experiment_count);
}


static void
gsl_error_handler (const char * reason,
		   const char * file,
		   int line,
		   int gsl_errno)
{
  std::ostringstream ostr;
  ostr << file
       << ": "
       << line
       << ": "
       << reason
       << " (GSL error code "
       << gsl_errno
       << ")";
  throw std::runtime_error(ostr.str());
}


TTest:: TTest(const Statistics &stats_a, const Statistics &stats_b)
  : _t(0.0),
    _df( stats_a._experiment_count + stats_b._experiment_count - 2 ),
    _p(0.0)
{
  // from http://www.socialresearchmethods.net/kb/stat_t.php
  /*  Rational term_a(stats_a._experiment_position_variance / stats_a._experiment_position_mean);
  Rational term_b(stats_b._experiment_position_variance / stats_b._experiment_position_mean);
  Real std_error( sqrt(static_cast<Real>(term_a + term_b)) );
  
  Rational mean_difference(stats_a._experiment_position_mean - stats_b._experiment_position_mean);
  Rational abs_mean_difference(abs(mean_difference));
  
  _t = abs_mean_difference / std_error;*/
  
  Rational mean_difference( stats_a._experiment_position_mean - stats_b._experiment_position_mean );
  
  Rational sqrt_dividend( (stats_a._experiment_count-1)*stats_a._experiment_position_variance
			  + (stats_b._experiment_count-1)*stats_b._experiment_position_variance );
  Rational sqrt_divisor(_df);
  Rational sqrt_second_term( static_cast<Rational>(1)/stats_a._experiment_count
			     + static_cast<Rational>(1)/stats_b._experiment_count );
  Real std_error( sqrt(static_cast<Real>( (sqrt_dividend/sqrt_divisor) * sqrt_second_term )) );
  
  _t = mean_difference / std_error;
  
  // from http://www.nrbook.com/a/bookcpdf/c6-4.pdf (my god what a trek)
  gsl_set_error_handler(gsl_error_handler);
  unsigned long v(_df.get_ui());
  _p = gsl_sf_beta_inc(0.5*v, 0.5, v / (v + static_cast<mpf_class>(_t*_t).get_d()));
}

Analyser:: Analyser(Db::Database &database,
		    unsigned trace_count,
		    const String &property_expression)
{
  Db::IntentInfoList intent_codes(database.get_intent_options(trace_count));
  Db::IdList experiments;
  
  String algo_spec;
  //  database.get_experiment_group_parameters(experiment_group, &algo_spec);
  Algorithm algo(algo_spec);
  _offset_probabilities = algo.get_probabilities();
  
  for (Db::IntentInfoList::iterator iter(intent_codes.begin());
       iter != intent_codes.end();
       ++iter)
    {
      experiments = database.get_experiments("[trace count] == 1 && [algorithm spec] == [1 1] && ( [bit count] > 50 || time <= 60 )");
      _statistic_sets.insert(StatisticSets::value_type(iter->first, Statistics(database, _offset_probabilities, experiments)));
    }
}

